package index.structures;

import java.util.Arrays;
import java.util.HashMap;

import javax.swing.JFrame;

import testapp.JOptionDialog;

import index.base.AIndexStructure;

public final class IndexAesa  extends AIndexStructure{
	
	final int[][] PRE_COMPUTED_NEIGHBORS;
	final int NUM_PRE_COMPUTED_NEIGHBORS = 10;
	int[][] tempDataStore;

	protected IndexAesa(int _size, int _dim, int _minValue, int _maxValue,
			HashMap<String, Object> _map) {
		super(_size, _dim, _minValue, _maxValue, _map);
		initParameter(_map);
		
		this.PRE_COMPUTED_NEIGHBORS = new int[_size][NUM_PRE_COMPUTED_NEIGHBORS];
		this.tempDataStore = new int[_size][];
	}

	int numPointsInserted = 0;
	@Override
	protected boolean insert(int[] _dataset) {
		tempDataStore[numPointsInserted++] = _dataset;
		if(numPointsInserted==this.size)
			computeNeighbors();
		return true;
	}

	/**
	 * for all
	 */
	private final void computeNeighbors() {
		for(int TID=0;TID<this.size;TID++){
			//TODO hier paralllelisieren
			if(TID%100==0)
				System.out.println(TID);
			computeNeighbors(TID,this.tempDataStore,this.PRE_COMPUTED_NEIGHBORS[TID], NUM_PRE_COMPUTED_NEIGHBORS);
		}
		//clean up
		tempDataStore = null;
	}

	private void computeNeighbors(final int QUERY_TID, final int[][] POINTS, final int[] RESULT, final int NUM) {
		final long[] distances 	= new long[size];
		final int[] query 		= POINTS[QUERY_TID];
		
		for(int pointTID=0;pointTID<size;pointTID++){
			distances[pointTID] = compDistSquared(query, POINTS[pointTID], pointTID);
		}
		Arrays.sort(distances);//TODO optimize me
		int resultTID = -1;
		for(int i=0;i<NUM_PRE_COMPUTED_NEIGHBORS;i++){
			resultTID = (int) (distances[i] & 0xFFFFFFFF);
			RESULT[i] = resultTID;
		}
	}
	
	final int compDistSquared(final int[] query, final int[] p2){
		int sum  = 0;
		int temp = 0;
		for(int dim=0;dim<query.length;dim++){
			temp =  query[dim]-p2[dim];
			sum  += (temp*temp);
		}
		return sum;
	}
	
	final long compDistSquared(final int[] query, final int[] p2, final int TID){
		int sum  = 0;
		int temp = 0;
		for(int dim=0;dim<query.length;dim++){
			temp =  query[dim]-p2[dim];
			sum  += (temp*temp);
		}
		long result = sum;
		result = (result << 32);
		result |= TID;
		return result;
	}

	@Override
	public int[] search(int[] _datasetToSearch) {
		for (int i = 0; i < size; i++) {
            boolean isIdentity = true;
            int[] test = getDataset(i);
            for (int k = 0; k < dim; k++) {
                if (test[k] != _datasetToSearch[k]) {
                    isIdentity = false;
                    break;
                }
            }
            if (isIdentity == true) {
                return test;
            }
        }
        return null;
	}

	final int NOT_FOUND = -1;
	final int getTID(int[] _datasetToSearch){
		for (int tid = 0; tid < size; tid++) {
            boolean isIdentity = true;
            int[] test = getDataset(tid);
            int curDim = 0;
            for (; curDim < dim; curDim++) {
                if (test[curDim] != _datasetToSearch[curDim]) {
                    isIdentity = false;
                    break;
                }
            }
            if (isIdentity == true) {
                return tid;
            }
        }
		return NOT_FOUND;
	}
	
	@Override
	public int[][] searchKNN(int[] _datasetToSearch, int _k) {
		if(_k>NUM_PRE_COMPUTED_NEIGHBORS)
			return null;
		final int TID = getTID(_datasetToSearch);
		if(TID == NOT_FOUND)
			return null;
		
		final int NUM_RESULTS = Math.min(_k, this.NUM_PRE_COMPUTED_NEIGHBORS);
		int[][] result = new int[NUM_RESULTS][];
		int nextNeighborTID = -1;
		int[] nextNeighbor;
		
		for(int i=0;i<NUM_RESULTS;i++){
			nextNeighborTID = this.PRE_COMPUTED_NEIGHBORS[TID][i];
			nextNeighbor = getDataset(nextNeighborTID);
			result[i] = nextNeighbor;
		}

		return result;
	}

	@Override
	public boolean canKNNSearch() {
		return true;
	}

	@Override
	public String getInformations() {
		return "AESA full with seq. scan";
	}

	@Override
	public String getName() {
		return "AESA";
	}

	@Override
	public void clear() {
		// TODO Auto-generated method stub
		
	}

	@Override
	protected JOptionDialog getOptionsDialog(JFrame _parent) {
		return null;
	}
	
	///////////////////////////////////////////TEST STUFF//////////////////////////////////////
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		int[][] data = getData();
		
		IndexAesa index = new IndexAesa(data.length, data[0].length, 0, 255, null);
		
		for(int point=0;point<data.length;point++){
			index.insert_Internal(data[point]);
		}

	}
	
	private static int[][] getData() {
		final int UPPER_BOUND = 200;
		final int[][] data = new int[UPPER_BOUND*UPPER_BOUND][2];
		System.out.println("Size of dataset "+UPPER_BOUND*UPPER_BOUND);
		int counter = 0;
		for(int x=0;x<UPPER_BOUND;x++){
			for(int y=0;y<UPPER_BOUND;y++){
				data[counter][0] = x;
				data[counter][1] = UPPER_BOUND-y;
				counter++;
			}
		}

		return data;
	}
}
